#region usings

using System;
using System.ComponentModel.Composition;
using System.Runtime.InteropServices;

using SlimDX;
using SlimDX.Direct3D9;
using VVVV.Core.Logging;
using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;
using VVVV.PluginInterfaces.V2.EX9;
using VVVV.Utils.VColor;
using VVVV.Utils.VMath;
using VVVV.Utils.SlimDX;

using xn;
using System.Collections.Generic;
using System.Threading;

#endregion usings

//here you can change the vertex type
using VertexType = VVVV.Utils.SlimDX.TexturedVertex;

namespace VVVV.Nodes
{
    #region PluginInfo
    [PluginInfo(Name = "Skeleton", Category = "OpenNI", Help = "OpenNi MultiSkeleton Reader", Tags = "plugin, OpenNI,Kinect")]
    #endregion PluginInfo
    public class OpenNISkeletonNode : DXTextureOutPluginBase, IPluginEvaluate
    {
        #region fields & pins

        [Input("Wave Count", DefaultValue = 1)]
        IDiffSpread<double> FWaveCountIn;

        [Input("Width", DefaultValue = 64)]
        IDiffSpread<int> FWidthIn;

        [Input("Height", DefaultValue = 64)]
        IDiffSpread<int> FHeightIn;

        [Input("ConfidenceTS", DefaultValue = 0.5)]
        ISpread<double> FConfidenceTSIn;

        [Output("Message")]
        ISpread<string> FMessage;

        [Output("TrackedUsersId")]
        ISpread<int> FTrackedUsersId;

        [Output("JointsPosition")]
        ISpread<ISpread<Vector3D>> FJointsPosition;

        [Import()]
        ILogger FLogger;

        //track the current texture slice
        int FCurrentSlice;

        #endregion fields & pins
        #region Kinect vars

        //public readonly string SAMPLE_XML_FILE = @"D:\developpa\KINECT\VVVV\Plugin0KinectTemplate/bin/Data/SamplesConfig.xml";
        public readonly string SAMPLE_XML_FILE = @"Plugins/OpenNISkeleton/Data/SamplesConfig.xml";

        private Context context;
        private xn.DepthGenerator depth;
        private UserGenerator userGenerator;
        private SkeletonCapability skeletonCapbility;
        private PoseDetectionCapability poseDetectionCapability;
        private string calibPose;
        private Thread readerThread;
        private bool shouldRun;
        // private Bitmap bitmap;
        private int[] histogram;

        private Dictionary<uint, Dictionary<SkeletonJoint, SkeletonJointPosition>> joints;

        private bool shouldDrawPixels = true;
        private bool shouldDrawBackground = true;
        private bool shouldPrintID = true;
        private bool shouldPrintState = true;
        private bool shouldDrawSkeleton = true;

        private bool kinectStarted = false;

        private List<uint> trackedUsers;


        #endregion

        // import host and hand it to base constructor
        [ImportingConstructor()]
        public OpenNISkeletonNode(IPluginHost host)
            : base(host)
        {
        }


        #region KINECT FUNCTIONZ
        private void getSkeleton(uint user, int userIndex)
        {

            GetJoints(user);
            Dictionary<SkeletonJoint, SkeletonJointPosition> dict = this.joints[user];

            FJointsPosition[userIndex].SliceCount = dict.Count;

            int jointIndex = 0;

            foreach (KeyValuePair<SkeletonJoint, SkeletonJointPosition> kvp in dict)
            {
                SkeletonJointPosition obj = kvp.Value;
               
               if(obj.fConfidence>=FConfidenceTSIn[0])
               {
                FJointsPosition[userIndex][jointIndex] = new Vector3D(obj.position.X, obj.position.Y, obj.position.Z);
			   }
                jointIndex++;
            }
        }
        public void initKinect()
        {

            kinectStarted = true;
			
			FMessage[0] = "Init OpenNI Node";
			
            this.context = new Context(SAMPLE_XML_FILE);
            this.depth = context.FindExistingNode(xn.NodeType.Depth) as DepthGenerator;

            if (this.depth == null)
            {
                //throw new Exception("Viewer must have a depth node!");
                FLogger.Log(LogType.Debug, "non ci siamo eh");
            }
            else
            {
                FLogger.Log(LogType.Debug, "Ecchime sono un Primesense attivato e running....:P");
            }

            this.userGenerator = new UserGenerator(this.context);
            this.skeletonCapbility = new SkeletonCapability(this.userGenerator);
            this.poseDetectionCapability = new PoseDetectionCapability(this.userGenerator);
            this.calibPose = this.skeletonCapbility.GetCalibrationPose();

            this.userGenerator.NewUser += new UserGenerator.NewUserHandler(userGenerator_NewUser);
            this.userGenerator.LostUser += new UserGenerator.LostUserHandler(userGenerator_LostUser);
            this.poseDetectionCapability.PoseDetected += new PoseDetectionCapability.PoseDetectedHandler(poseDetectionCapability_PoseDetected);
            this.skeletonCapbility.CalibrationEnd += new SkeletonCapability.CalibrationEndHandler(skeletonCapbility_CalibrationEnd);

            this.skeletonCapbility.SetSkeletonProfile(SkeletonProfile.All);
            this.joints = new Dictionary<uint, Dictionary<SkeletonJoint, SkeletonJointPosition>>();
            this.userGenerator.StartGenerating();

            FMessage[0] = "OpenNI Node registered";

            this.histogram = new int[this.depth.GetDeviceMaxDepth()];

            MapOutputMode mapMode = this.depth.GetMapOutputMode();

            //this.bitmap = new Bitmap((int)mapMode.nXRes, (int)mapMode.nYRes/*, System.Drawing.Imaging.PixelFormat.Format24bppRgb*/);

            this.shouldRun = true;
            this.readerThread = new Thread(ReaderThread);
            this.readerThread.Start();
            
            FMessage[0] = "OpenNI Thread Started ";

        }

        private unsafe void CalcHist(DepthMetaData depthMD)
        {
            // reset
            for (int i = 0; i < this.histogram.Length; ++i)
                this.histogram[i] = 0;

            ushort* pDepth = (ushort*)depthMD.DepthMapPtr.ToPointer();

            int points = 0;
            for (int y = 0; y < 480; ++y)
            {
                for (int x = 0; x < 640; ++x, ++pDepth)
                {
                    ushort depthVal = *pDepth;
                    if (depthVal != 0)
                    {
                        this.histogram[depthVal]++;
                        points++;
                    }
                }
            }

            for (int i = 1; i < this.histogram.Length; i++)
            {
                this.histogram[i] += this.histogram[i - 1];
            }

            if (points > 0)
            {
                for (int i = 1; i < this.histogram.Length; i++)
                {
                    this.histogram[i] = (int)(256 * (1.0f - (this.histogram[i] / (float)points)));
                }
            }
        }

        private unsafe void ReaderThread()
        {

            //DepthMetaData depthMD = new DepthMetaData();

            while (this.shouldRun)
            {
                /*try
                {
                    this.context.WaitOneUpdateAll(this.depth);
                }
                catch (Exception)
                {
                }*/

                //this.depth.GetMetaData(depthMD);

                //CalcHist(depthMD);

                lock (this)
                {
                    #region drawPixels
                    /*
                        //Rectangle rect = new Rectangle(0, 0, this.bitmap.Width, this.bitmap.Height);
                        //BitmapData data = this.bitmap.LockBits(rect, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format24bppRgb);
					
                       
                    if (this.shouldDrawPixels)
                    {
                        ushort* pDepth = (ushort*)this.depth.GetDepthMapPtr().ToPointer();
                        ushort* pLabels = (ushort*)this.userGenerator.GetUserPixels(0).SceneMapPtr.ToPointer();

                        // set pixels
                        for (int y = 0; y < depthMD.YRes; ++y)
                        {
                            byte* pDest = (byte*)data.Scan0.ToPointer() + y * data.Stride;
                            for (int x = 0; x < depthMD.XRes; ++x, ++pDepth, ++pLabels, pDest += 3)
                            {
                                pDest[0] = pDest[1] = pDest[2] = 0;

                                ushort label = *pLabels;
                                if (this.shouldDrawBackground || *pLabels != 0)
                                {
                                    Color labelColor = Color.White;
                                    if (label != 0)
                                    {
                                        labelColor = colors[label % ncolors];
                                    }

                                    byte pixel = (byte)this.histogram[*pDepth];
                                    pDest[0] = (byte)(pixel * (labelColor.B / 256.0));
                                    pDest[1] = (byte)(pixel * (labelColor.G / 256.0));
                                    pDest[2] = (byte)(pixel * (labelColor.R / 256.0));
                                }
                            }
                        }*/
                    #endregion
                }

                //this.bitmap.UnlockBits(data);
                //Graphics g = Graphics.FromImage(this.bitmap);

                uint[] users = this.userGenerator.GetUsers();
                trackedUsers = new List<uint>();

                foreach (uint user in users)
                {
                    #region labels
                    /*
                        if (this.shouldPrintID)
                        {
                            Point3D com = this.userGenerator.GetCoM(user);
                            com = this.depth.ConvertRealWorldToProjective(com);

                            string label = "";
                            if (!this.shouldPrintState)
                                label += user;
                            else if (this.skeletonCapbility.IsTracking(user))
                                label += user + " - Tracking";
                            else if (this.skeletonCapbility.IsCalibrating(user))
                                label += user + " - Calibrating...";
                            else
                                label += user + " - Looking for pose";

                            g.DrawString(label, new Font("Arial", 6), new SolidBrush(anticolors[user % ncolors]), com.X, com.Y);

                        } */
                    #endregion


                    if (this.skeletonCapbility.IsTracking(user))
                    {
                        trackedUsers.Add(user);

                    }
                }
                int trackedUsersCount = trackedUsers.Count;
                /// tracked user ID
                FTrackedUsersId.SliceCount = trackedUsersCount;

                for (int i = 0; i < trackedUsersCount; i++)
                {
                    FTrackedUsersId[i] = (int)trackedUsers[i];
                }
                
                //////////////  Position spreading count
                FJointsPosition.SliceCount = trackedUsersCount;

                int index = 0;

                foreach (uint trackedUser in trackedUsers)
                {
                    getSkeleton(trackedUser, index);
                    index++;
                }
            }
        }


        private void GetJoint(uint user, SkeletonJoint joint)
        {
            SkeletonJointPosition pos = new SkeletonJointPosition();
            this.skeletonCapbility.GetSkeletonJointPosition(user, joint, ref pos);

            if (pos.position.Z == 0)
            {
                pos.fConfidence = 0;
            }
            else
            {
                pos.position = this.depth.ConvertRealWorldToProjective(pos.position);
            }
            this.joints[user][joint] = pos;

        }

        private void GetJoints(uint user)
        {
            GetJoint(user, SkeletonJoint.Head);
            GetJoint(user, SkeletonJoint.Neck);

            GetJoint(user, SkeletonJoint.LeftShoulder);
            GetJoint(user, SkeletonJoint.LeftElbow);
            GetJoint(user, SkeletonJoint.LeftHand);

            GetJoint(user, SkeletonJoint.RightShoulder);
            GetJoint(user, SkeletonJoint.RightElbow);
            GetJoint(user, SkeletonJoint.RightHand);

            GetJoint(user, SkeletonJoint.Torso);

            GetJoint(user, SkeletonJoint.LeftHip);
            GetJoint(user, SkeletonJoint.LeftKnee);
            GetJoint(user, SkeletonJoint.LeftFoot);

            GetJoint(user, SkeletonJoint.RightHip);
            GetJoint(user, SkeletonJoint.RightKnee);
            GetJoint(user, SkeletonJoint.RightFoot);
        }

        void skeletonCapbility_CalibrationEnd(ProductionNode node, uint id, bool success)
        {
            if (success)
            {
                this.skeletonCapbility.StartTracking(id);
                this.joints.Add(id, new Dictionary<SkeletonJoint, SkeletonJointPosition>());

                FMessage[0] = "Calibration ended for user " + id.ToString();
            }
            else
            {
                this.poseDetectionCapability.StartPoseDetection(calibPose, id);
                FMessage[0] = "Pose detection for user  " + id.ToString();
            }
        }

        void poseDetectionCapability_PoseDetected(ProductionNode node, string pose, uint id)
        {
            this.poseDetectionCapability.StopPoseDetection(id);
            this.skeletonCapbility.RequestCalibration(id, true);

            FMessage[0] = "Pose detected " + id.ToString();
        }

        void userGenerator_NewUser(ProductionNode node, uint id)
        {
            this.poseDetectionCapability.StartPoseDetection(this.calibPose, id);

            FMessage[0] = "New user" + id.ToString();
        }

        void userGenerator_LostUser(ProductionNode node, uint id)
        {
            this.joints.Remove(id);

            FMessage[0] = "Lost user  " + id.ToString();
        }
        #endregion

        //called when data for any output pin is requested
        unsafe public void Evaluate(int SpreadMax)
        {
            SetSliceCount(SpreadMax);

            if (!kinectStarted)
            {
                initKinect();
            }

            //recreate texture if resolution was changed
            if (FWidthIn.IsChanged || FHeightIn.IsChanged)
            {
                //set new texture size
                Reinitialize();
            }

            //update texture
            //if (FWaveCountIn.IsChanged)
            //{
            try
                {
                    this.context.WaitOneUpdateAll(this.depth);
                }
                catch (Exception)
                {
                }
            Update();
            //}


        }

        //this method gets called, when Reinitialize() was called in evaluate,
        //or a graphics device asks for its data
        protected override Texture CreateTexture(int Slice, SlimDX.Direct3D9.Device device)
        {
            FLogger.Log(LogType.Debug, "Creating new texture at slice: " + Slice);
            return TextureUtils.CreateTexture(device, 640, 480);
        }

        //this method gets called, when Update() was called in evaluate,
        //or a graphics device asks for its texture, here you fill the texture with the actual data
        //this is called for each renderer, careful here with multiscreen setups, in that case
        //calculate the pixels in evaluate and just copy the data to the device texture here
        unsafe protected override void UpdateTexture(int Slice, Texture texture)
        {
        	
            FCurrentSlice = Slice;
            DepthMetaData depthMD = new DepthMetaData();
			this.depth.GetMetaData(depthMD);

            CalcHist(depthMD);
            /*ushort* pDepth = (ushort*)this.depth.GetDepthMapPtr().ToPointer();
            var rect = texture.LockRectangle(0, LockFlags.None).Data; 
            for(int y = 0; y > 480; y++) 
            {
                for(int x = 0; x > 640; x++, pDepth++)
                {
                    byte depthPixel = (byte)this.histogram[*pDepth];
                    uint pixColor = UInt32Utils.fromARGB(255,255,depthPixel,depthPixel);
                    rect.Write(pixColor); 
                }
            }
            texture.UnlockRectangle(0);*/

            //ushort* pDepth = (ushort*)this.depth.GetDepthMapPtr().ToPointer();
            TextureUtils.Fill32BitTexInPlace(texture, FillTexure);
        }

        //this is a pixelshader like method, which we pass to the fill function
        unsafe public void FillTexure(uint* data, int row, int col, int width, int height)
        {
            
            ushort* pDepth = (ushort*)this.depth.GetDepthMapPtr().ToPointer();
            ushort* pix = pDepth+col+row*width;
			byte depthPixel = (byte)this.histogram[*pix];
            //a pixel is just a 32-bit unsigned int value
            var pixel = UInt32Utils.fromARGB(255, depthPixel, depthPixel, depthPixel);

            //copy pixel into texture
            TextureUtils.SetPtrVal2D(data, pixel, row, col, width);
        }
    }
}
